#include "StatAlgorithm.h"
using namespace std;

// Inverse CDF of normal. Input: probability p		
// Accuracy: about 5 decimial points
double norminv(double p)
{
	const double PLIM = 1.0E-18, P0 = -0.322232431088E0, P1 = -1.0, P2 = -0.342242088547E0;
	const double P3 = -0.0204231210245E0, P4 = -0.453642210148E-4, Q0 = 0.099348462606E0, \
		Q1 = 0.588581570495E0, Q2 = 0.531103462366E0, Q3 = 0.10353775285E0, Q4 = 0.38560700634E-2;
	double vtemp;
	require(0 < p && p < 1, "Invalid probability for normal-inverse function");
	if (0.5 < p) p = 1 - p;
	if (PLIM <= p) 
	{
		double y = sqrt(-log(p*p));
		vtemp = y + ((((y*P4 + P3)*y + P2)*y + P1)*y +P0)/((((y*Q4 + Q3)*y + Q2)*y + Q1)*y + Q0);
	}
	else
	{
		// This is as far out in the tails as we go
		vtemp = 8;
	}
	if (0.5 <= p) vtemp = -vtemp;
	return vtemp; 
}
/* Inverse CDF of student-t distribution. Adapted from A guide to simulation (2nd ed) by Paul Bratley
 * pp. 340
 * Inputs: 	p = probability, df = degree of freedom
 * Accuracy in decimal digits:  5 if d.f. >= 8, machine precision if df = 1 or 2, 3 otherwise
 */
double tinv(double p, int df)
{
	const double DWARF = 0.1E-30, PIOVR2 = 1.5707963268E0;
	double t;
	double w[4];
	require(0 < p && p < 1, "Invalid probability for t-inverse function");
	require(df >= 0, "Invalid DF for t-inverse function");
	if (df == 0)
	{ // This may happen with a deterministic simulator, then simply return 0
		return 0;
	}
	// if df = 1, it is a Cauchy, use special method (exact)
	if (df == 1)
	{
		double arg = 3.1415926 * (1 - p);
		double sinarg = (DWARF < sin(arg)) ? sin(arg) : DWARF;
		return cos(arg) / sinarg;
	}
	// if df ==2, use a special method
	if (df == 2)
	{
		t = (p <= 0.5) ? (2 * p) :  (2 * (1 - p));
		if (t <= DWARF) t = DWARF;
		double vtemp = sqrt((2/(t*(2 - t))) - 2);
		if (p <= 0.5) vtemp = -vtemp;
		return vtemp;
	}	
	// Check to see if we are too far out in the tails
	if (DWARF < p && DWARF < (1 -p))
	{// General case
		double px2 = 2 * p;
		double p2tail = (px2 < (2 - px2)) ? px2 : 2 - px2;
		//begin the approximiation. 
		w[1] = 1.0 / (df - 0.5);
		w[0] = 48 / (w[1]*w[1]);
		w[2] = ((20700*w[1]/w[0] - 98)*w[1] -16)*w[1] + 96.36;
		w[3] = ((94.5/(w[0] + w[2]) - 3)/w[0] + 1)*sqrt(w[1]*PIOVR2)*df;
		double x = w[3] * p2tail;
		double c = w[2];
		double y = pow(x, 2.0/double(df));
		if(y <= (0.05 + w[1])) 
		{
			double y2 = (df + 6)/double(df + y) - 0.089*w[3] - 0.822;
			y2 *= (df + 2) * 3;
			y2 = (1/y2 + .5/(df + .4))*y - 1;
			y2 *= double(df + 1)/(df + 2);
			y = y2 + 1/y;
		}
		else
		{
			//asymptotic inverse expansion about normal
			x = norminv(p2tail * .5);
			y = x * x;
			if (df < 5) c += 0.3 * (df - 4.5) * (x + .6);
			c += (((.05*w[3]*x - 5)*x - 7)*x - 2)*x + w[0];
			y = (((((.4*y + 6.3)*y + 36)*y + 94.5)/c - y - 3)/w[0] + 1)*x;
			y = w[1] * y * y;
			if (y <= .002) 
			{
				y = .5 * y * y + y;
			}
			else
			{
				y = exp(y) - 1;
			}
		}
		t = sqrt(df * y);
	}
	else
	{// Tails
		t = 10E30;
	}
	if (p < 0.5) t = -t;
	return t;
}  		
